#cython: language_level=3
"""
Python binding for jq
"""

import os


class ScriptRuntimeError(Exception):
    """
    Exception thrown when a script calls error()
    """
    pass


cdef extern from "jv.h":
    ctypedef struct jv:
        pass
    const char* jv_string_value(jv)

    ctypedef enum jv_kind:
      JV_KIND_INVALID,
      JV_KIND_NULL,
      JV_KIND_FALSE,
      JV_KIND_TRUE,
      JV_KIND_NUMBER,
      JV_KIND_STRING,
      JV_KIND_ARRAY,
      JV_KIND_OBJECT


    jv jv_copy(jv)

    jv_kind jv_get_kind(jv)
    jv jv_invalid_get_msg(jv)
    int jv_invalid_has_msg(jv)

    void jv_free(jv)

    jv jv_null()
    jv jv_bool(int)

    jv jv_number(double)
    double jv_number_value(jv)
    bint jv_is_integer(jv)

    jv jv_array()
    jv jv_array_sized(int)
    int jv_array_length(jv)
    jv jv_array_get(jv, int)
    jv jv_array_set(jv, int, jv)
    jv jv_array_append(jv, jv)

    jv jv_string_sized(const char*, int)
    jv jv_dump_string(jv, int) 

    jv jv_object()
    jv jv_object_get(jv object, jv key)
    jv jv_object_set(jv object, jv key, jv value)

    int jv_object_iter(jv)
    int jv_object_iter_next(jv, int)
    int jv_object_iter_valid(jv, int)
    jv jv_object_iter_key(jv, int)
    jv jv_object_iter_value(jv, int)

    ctypedef enum jv_parser_flags:
      JV_PARSE_EXPLODE_TOPLEVEL_ARRAY

    cdef struct jv_parser:
        pass

    jv_parser* jv_parser_new(jv_parser_flags)
    void jv_parser_free(jv_parser*)
    void jv_parser_set_buf(jv_parser*, const char*, int, int)
    jv jv_parser_next(jv_parser*)


cdef extern from "jq.h":
    ctypedef struct jq_state:
        pass

    ctypedef void (*jq_err_cb)(void *, jv)

    jq_state *jq_init()
    void jq_set_attr(jq_state *, jv, jv)
    void jq_teardown(jq_state **)
    bint jq_compile_args(jq_state *, const char* str, jv args)
    void jq_start(jq_state *, jv value, int flags)
    jv jq_next(jq_state *)
    void jq_set_error_cb(jq_state *, jq_err_cb, void *)


cdef object jv_to_pyobj(jv jval):
    kind = jv_get_kind(jval)

    if kind == JV_KIND_NULL:
        return None
    elif kind == JV_KIND_FALSE:
        return False
    elif kind == JV_KIND_TRUE:
        return True
    elif kind == JV_KIND_NUMBER:
        v = jv_number_value(jval)
        if jv_is_integer(jval):
            return int(v)
        return v
    elif kind == JV_KIND_STRING:
        return jv_string_value(jval).decode('utf-8')
    elif kind == JV_KIND_ARRAY:
        alist = []
        for i in range(jv_array_length(jv_copy(jval))):
            value = jv_array_get(jv_copy(jval), i)
            alist.append(jv_to_pyobj(value))
            jv_free(value)
        return alist
    elif kind == JV_KIND_OBJECT:
        adict = {}
        it = jv_object_iter(jval)
        while jv_object_iter_valid(jval, it):
            key = jv_object_iter_key(jval, it)
            k = jv_to_pyobj(key)
            jv_free(key)
            value = jv_object_iter_value(jval, it)
            v = jv_to_pyobj(value)
            jv_free(value)
            adict[k] = v
            it = jv_object_iter_next(jval, it)
        return adict


cdef jv pyobj_to_jv(object pyobj) except *:
    if isinstance(pyobj, str):
        pyobj = pyobj.encode('utf-8')
        return jv_string_sized(pyobj, len(pyobj))
    elif isinstance(pyobj, bytes):
        return jv_string_sized(pyobj, len(pyobj))
    elif isinstance(pyobj, bool):
        return jv_bool(pyobj)
    elif isinstance(pyobj, (int, long, float)):
        return jv_number(pyobj)
    elif isinstance(pyobj, (list, tuple)):
        jval = jv_array()
        for i, item in enumerate(pyobj):
            jval = jv_array_append(jval, pyobj_to_jv(item))
        return jval
    elif isinstance(pyobj, dict):
        jval = jv_object()
        for key, value in pyobj.items():
            if not isinstance(key, str):
                raise TypeError("Key of json object must be a str, but got {}".format(type(key)))
            jval = jv_object_set(jval, pyobj_to_jv(key), pyobj_to_jv(value))
        return jval
    elif pyobj is None:
        return jv_null()
    else:
        raise TypeError("{!r} could not be converted to json".format(type(pyobj)))


cdef void Script_error_cb(void* x, jv err):
    Script._error_cb(<object>x, err)


cdef class Script:
    'Compiled jq script object'
    cdef object _errors
    cdef jq_state* _jq

    def __init__(self, const char* script, vars={}, library_paths=[]):
        self._errors = []
        self._jq = jq_init()
        if not self._jq:
            raise RuntimeError('Failed to initialize jq')
        jq_set_error_cb(self._jq, Script_error_cb, <void*>self)

        args = pyobj_to_jv([
            dict(name=k, value=v)
            for k, v in vars.items()
        ])

        jq_set_attr(
            self._jq,
            pyobj_to_jv("JQ_LIBRARY_PATH"),
            pyobj_to_jv([str(path) for path in library_paths])
        )

        if not jq_compile_args(self._jq, script, args):
            raise ValueError("\n".join(self._errors))

    cdef _error_cb(self, jv err):
        self._errors.append(jv_string_value(err).decode('utf-8'))

    def __dealloc__(self):
        jq_teardown(&self._jq)

    def all(self, pyobj):
        "Transform object by jq script, returning all results as list"
        cdef jv value = pyobj_to_jv(pyobj)
        jq_start(self._jq, value, 0)
        cdef list output = []

        while True:
            result = jq_next(self._jq)
            try:
                kind = jv_get_kind(result)
                if kind == JV_KIND_INVALID:
                    if not jv_invalid_has_msg(jv_copy(result)):
                        break
                    m = jv_invalid_get_msg(jv_copy(result))
                    e = str(jv_to_pyobj(m))
                    raise ScriptRuntimeError(e)
                else:
                    output.append(jv_to_pyobj(result))
            finally:
                jv_free(result)
        return output

    apply = all

    def first(self, value, default=None):
        """
        Transform object by jq script, returning the first result.
        Return default if result is empty.
        """
        ret = self.apply(value)
        if not ret:
            return default
        return ret[0]

    def one(self, value):
        """
        Transform object by jq script, returning the first result.
        Raise IndexError unless results does not include exactly one element.
        """
        ret = self.apply(value)
        if not ret:
            raise IndexError("Result of jq is empty")
        elif len(ret) > 1:
            raise IndexError("Result of jq have multiple elements")
        return ret[0]



# encoding: utf8
from __future__ import unicode_literals

import io
import os.path
import re
import shutil
import tempfile
import unittest
from datetime import datetime
from unittest.mock import patch

import pytest

import _pyjq
import pyjq


def test_compile_dot():
    s = pyjq.compile(".")
    assert isinstance(s, _pyjq.Script)


def test_syntax_error():
    with pytest.raises(ValueError, match=r"error: syntax error"):
        pyjq.compile("**")


def test_non_json_data():
    with pytest.raises(TypeError, match=re.escape("<class 'datetime.datetime'> could not be converted to json")):
        pyjq.all(".", {"date": datetime.now()})


def test_conversion_between_python_object_and_jv():
    objects = [
        None,
        False,
        True,
        1,
        1.5,
        "string",
        [None, False, True, 1, 1.5, [None, False, True], {"foo": "bar"}],
        {
            "key1": None,
            "key2": False,
            "key3": True,
            "key4": 1,
            "key5": 1.5,
            "key6": [None, False, True, 1, 1.5, [None, False, True], {"foo": "bar"}],
        },
    ]

    s = pyjq.compile(".")
    for obj in objects:
        assert [obj] == s.all(obj)


def test_assigning_values():
    assert pyjq.one("$foo", {}, vars=dict(foo="bar")) == "bar"
    assert pyjq.one("$foo", {}, vars=dict(foo=["bar"])) == ["bar"]


def test_all():
    assert pyjq.all(".[] | . + $foo", ["val1", "val2"], vars=dict(foo="bar")) == ["val1bar", "val2bar"]
    assert pyjq.all(". + $foo", "val", vars=dict(foo="bar")) == ["valbar"]


def test_first():
    assert pyjq.first(".[] | . + $foo", ["val1", "val2"], vars=dict(foo="bar")) == "val1bar"


def test_one():
    assert pyjq.one(". + $foo", "val", vars=dict(foo="bar")) == "valbar"

    # if got multiple elements
    with pytest.raises(IndexError):
        pyjq.one(".[]", [1, 2])

    # if got no elements
    with pytest.raises(IndexError):
        pyjq.one(".[]", [])


def test_url_argument():
    class FakeResponse:
        def getheader(self, name):
            return "application/json;charset=SHIFT_JIS"

        def read(self):
            return '["Hello", "世界", "！"]'.encode("shift-jis")

    with patch("urllib.request.urlopen", return_value=FakeResponse()):
        assert pyjq.all(".[] | . + .", url="http://example.com") == ["HelloHello", "世界世界", "！！"]

    def opener(url):
        return [1, 2, 3]

    assert pyjq.all(".[] | . + .", url="http://example.com", opener=opener) == [2, 4, 6]


def test_library_path(tmp_path_factory):
    library_path = tmp_path_factory.mktemp("a")
    library_path2 = tmp_path_factory.mktemp("b")

    library_file = library_path / "greeting.jq"
    library_file2 = library_path2 / "increment.jq"

    with library_file.open("w", encoding="ascii") as f:
        f.write('def hello: "HELLO";')
        f.write('def world: "WORLD";')
    with library_file2.open("w", encoding="ascii") as f:
        f.write("def increment: . + 1;\n")
    values = pyjq.all(
        'include "greeting"; include "increment"; .[] | [. | increment, hello, world]',
        [1, 2, 3],
        library_paths=[str(library_path), library_path2],  # It accepts both of str and pathlib.Path
    )
    assert [[2, "HELLO", "WORLD"], [3, "HELLO", "WORLD"], [4, "HELLO", "WORLD"]] == values


def test_script_runtime_error_exported():
    pyjq.ScriptRuntimeError # exported


#!/usr/bin/env python
import io
import lib2to3.main
import os
import platform
import shutil
import subprocess
import sys
import sysconfig
import tarfile
from os.path import abspath, dirname, join
from pathlib import Path

from setuptools import setup
from setuptools.command.build_ext import _build_ext
from setuptools.extension import Extension

dependencies_dir_path = join(abspath(dirname(__file__)), "dependencies")

onig_tarball_path = join(dependencies_dir_path, "onig-6.9.0.tar.gz")
onig_install_path = join(dependencies_dir_path, "onig_install")
onig_source_path = join(dependencies_dir_path, "onig-6.9.0")

jq_tarball_path = join(dependencies_dir_path, "jq-1.5.tar.gz")
jq_install_path = join(dependencies_dir_path, "jq_install")
jq_source_path = join(dependencies_dir_path, "jq-jq-1.5")


class build_ext(_build_ext):
    def run(self):
        self._build_oniguruma()
        self._build_libjq()
        _build_ext.run(self)

    def _build_oniguruma(self):
        self._safe_rmtree(onig_install_path)
        self._safe_rmtree(onig_source_path)

        self._extract_tarball(onig_tarball_path, dependencies_dir_path)
        for f in Path(onig_source_path).glob('src/*.py'):
            lib2to3.main.main("lib2to3.fixes", args=['-w', '--no-diffs', '-f', 'print', str(f)])

        self._build_lib(
            lib_dir=onig_source_path,
            commands=[
                ["./configure", "CFLAGS=-fPIC", "--disable-shared", "--prefix", onig_install_path],
                ["make"],
                ["make", "install"],
            ]
        )

    def _build_libjq(self):
        self._safe_rmtree(jq_install_path)
        self._safe_rmtree(jq_source_path)

        self._extract_tarball(jq_tarball_path, dependencies_dir_path)
        self._build_lib(
            lib_dir=jq_source_path,
            commands=[
                ["autoreconf", "-i"],
                ["./configure", "CFLAGS=-fPIC", "--disable-maintainer-mode",
                 "--enable-all-static", "--disable-shared",
                 "--with-oniguruma=" + onig_install_path, "--prefix", jq_install_path],
                ["make", "install-libLTLIBRARIES", "install-includeHEADERS"],
            ]
        )

    def _build_lib(self, lib_dir, commands):
        macosx_deployment_target = sysconfig.get_config_var("MACOSX_DEPLOYMENT_TARGET")
        if macosx_deployment_target:
            os.environ['MACOSX_DEPLOYMENT_TARGET'] = macosx_deployment_target

        for command in commands:
            subprocess.check_call(command, cwd=lib_dir)

    def _extract_tarball(self, tarball_path, workdir_path):
        tarfile.open(tarball_path, "r:gz").extractall(workdir_path)

    def _safe_rmtree(self, d):
        try:
            shutil.rmtree(d)
        except OSError:
            pass

libraries = ["jq", "onig"]
if platform.architecture()[1] == 'WindowsPE':
    libraries.append("shlwapi")

pyjq = Extension(
    "_pyjq",
    sources=["_pyjq.c"],
    include_dirs=["dependencies/jq_install/include"],
    libraries=libraries,
    library_dirs=["dependencies/jq_install/lib", "dependencies/onig_install/lib"]
)

setup(
    ext_modules=[pyjq],
    cmdclass={"build_ext": build_ext},
    package_data={'': [onig_tarball_path, jq_tarball_path]},
)



import json
import re
import urllib

import _pyjq


ScriptRuntimeError = _pyjq.ScriptRuntimeError


def compile(script, vars={}, library_paths=[]):
    """
    Compile a jq script, retuning a script object.

    library_paths is a list of strings that defines the module search path.
    """

    return _pyjq.Script(script.encode("utf-8"), vars=vars, library_paths=library_paths)


def default_opener(url):
    res = urllib.request.urlopen(url)
    content_type = res.getheader("Content-Type")

    if content_type is None:
        encoding = "utf-8"
    else:
        m = re.search(r"charset\s*=\s*(\S+)", content_type, re.I)
        if m is None:
            encoding = "utf-8"
        else:
            encoding = m.group(1)
    return json.loads(res.read().decode(encoding))


def _get_value(value, url, opener):
    if url is not None:
        if value is not None:
            raise TypeError("'value' and 'url' are not able to specified at the same time")
        return opener(url)
    return value


def all(script, value=None, vars={}, url=None, opener=default_opener, library_paths=[]):
    """
    Transform value by script, returning all results as list.
    """
    return compile(script, vars, library_paths).all(_get_value(value, url, opener))


def apply(script, value=None, vars={}, url=None, opener=default_opener, library_paths=[]):
    """
    Transform value by script, returning all results as list.
    """
    return all(script, value, vars, url, opener, library_paths)


apply.__doc__ = all.__doc__


def first(script, value=None, default=None, vars={}, url=None, opener=default_opener, library_paths=[]):
    """
    Transform object by jq script, returning the first result.
    Return default if result is empty.
    """
    return compile(script, vars, library_paths).first(_get_value(value, url, opener), default)


def one(script, value=None, vars={}, url=None, opener=default_opener, library_paths=[]):
    """
    Transform object by jq script, returning the first result.
    Raise ValueError unless results does not include exactly one element.
    """
    return compile(script, vars, library_paths).one(_get_value(value, url, opener))
